#!/usr/bin/env python3
# LICENSE: MIT
# do whatever you like with this code, all at your own risk
# docs: https://www.ratrabbit.nl/ratrabbit/wdcms
from bs4 import BeautifulSoup
import copy
import datetime
import fnmatch
import markdown
import markdown2
import multiprocessing
import os
import re
import shutil
import sys
from subprocess import Popen, PIPE

VERSION = "0.9.9.1"

# ChangeLog
# 0.9.8   initial release
# 0.9.9   add encoding=self.encoding at every (F)open statement
#         define self.encoding as "utf-8"
#         add ChangeLog
#         add self.copyright
#         use shutil.copytree in stead of distutils.dir_util.copy_tree
#         add multiprocessing using os.fork() if available
# 0.9.9.1 add code to wait for end of all forked processes


class wdcms():  # OK, let us define a class ...

    def __init__(self, arg=None):
        self.mode                    = arg
        self.leftq                   = 'YYY'
        self.rightq                  = 'ZZZ'
        self.leftqq                  = self.leftq.lower()
        self.rightqq                 = self.rightq.lower()
        self.startparams             = {}
        self.startparams['css']      = []   # list of pevious .css files
        self.startparams['js']       = []   # list of previous .js files
        self.startparams['fname']    = ""   # file name
        self.startparams['urldir']   = ""   # current directory
        self.startparams['content']  = ""   # content
        self.startparams['level']    = 0    # recursion level
        self.startparams['template'] = ""   # template
        self.startparams['head']     = []   # list of addendums for <head>
        self.startparams['trans']    = {}   # translations
        self.startparams['settings'] = {}   # settings
        self.startparams['menu']     = []   # menu
        date = datetime.datetime.now()
        date = date.strftime("%Y/%m/%d %H:%M")  # alas, %Z shows nothing ...
        self.handle_translations(
            self.startparams, ("wdcms_date: "+date, "wdcms_version: "+VERSION, "leftq: "+self.leftqq, "rightq:"+self.rightqq))
        self.handle_settings(
                self.startparams, ("superfish: yes", "format: pandoc", "parser: html.parser"))

        self.numdirs     = 0
        self.nummd       = 0
        self.numcss      = 0
        self.numjs       = 0
        self.numhead     = 0
        self.numtrans    = 0
        self.numsettings = 0
        self.nummenu     = 0
        self.numcopy     = 0
        self.forked      = 0
        self.encoding    = "utf-8"
        self.copyright   = "<!-- Generated by wdcms version " + VERSION +  \
            " Copyright 2021, 2022 W. Vermin -->\n"

    def splitline(self, line, c):
        l = re.sub(r'#.*$', r'', line)
        l = l.split(c, 1)
        if len(l) == 2:
            return l[0].strip(), l[1].strip()
        return "", ""

    def doitdir(self, params): self.numdirs += 1

    def level2dir(self, level):
        d = '.'
        for i in range(level):
            d += '/..'
        return d

    def rep_(self, a): return a.replace("_", " ").strip()

    def esclinkname(self, l): return self.rep_(self.leftq+l+self.rightq)

    def linkname(self, l): return self.leftq+l+self.rightq

    def do_format(self, inputdata, params):
        formatter = params['settings']['format']
        data = inputdata
        a = data.split('\n', 1)
        if len(a) == 2:
            k, v = self.splitline(a[0], ':')
            if k == "format":
                formatter = v
                data = a[1]
        if formatter == "markdown":
            return markdown.markdown(data)
        elif formatter == "markdown2":
            return markdown2.markdown(data)
        elif formatter == "pandoc":
            p = Popen(["pandoc", "-f", "markdown-smart-auto_identifiers", "-t", "html"],
                      text=True, stdin=PIPE, stdout=PIPE, encoding=self.encoding)
            # -smart: disable smart quotes
            # -auto_identifiers: disable generation of id in <h> <h1> ...
            out, err = p.communicate(input=data)
            return out
        elif formatter == "mediawiki":
            p = Popen(["pandoc", "-f", "mediawiki-smart-auto_identifiers", "-t", "html"],
                      text=True, stdin=PIPE, stdout=PIPE, encoding=self.encoding)
            # -smart: disable smart quotes
            # -auto_identifiers: disable generation of id in <h> <h1> ...
            out, err = p.communicate(input=data)
            return out
        elif formatter == "asis":
            return data
        else:
            print("Unknown formatter defined in", params['urldir']+': ', formatter +
                  ". Valid formatters are 'pandoc' 'markdown' 'markdown2' 'asis' 'mediawiki'")
            sys.exit(1)

    class repattr:
        def has_attr(self, tag): return tag.has_attr(self.attr)

        def doit(self, to_base, soup, attr):
            self.attr = attr
            for child in soup.find_all(self.has_attr):
                x = child[attr]
                if len(x) >= 1 and x[0] == '/':
                    if len(x) >= 2 and x[1] == '/':
                        child[attr] = x[1:]
                    else:
                        child[attr] = to_base + x

    def handle_translations(self, params, lines):
        for line in lines:
            k, v = self.splitline(line, ':')
            if k:
                params['trans'][self.linkname(k)] = v
                if '_' in k:  # if so, use also the key with '_' replaced by ' '
                    params['trans'][self.esclinkname(k)] = v

    def handle_settings(self, params, lines):
        for line in lines:
            k, v = self.splitline(line, ':')
            if k:
                if k in ['format', 'superfish', 'parser']:
                    params['settings'][k] = v
                else:
                    print("Invalid settings in",
                          './'+params['urldir'], ": '"+k+"', quitting")
                    sys.exit(1)

    def handle_menu(self, params, lines):
        firstmenu = True
        for line in lines:
            k, v = self.splitline(line, ':')
            if k:
                if k == 'menu':
                    if firstmenu:
                        firstmenu = False
                        params['menu'] = []
                    l, w = self.splitline(v, ':')
                    if l:
                        params['menu'] += [(l, w)]

    def handle_content(self, params):
        content  = params['content']
        css      = params['css']
        head     = params['head']
        js       = params['js']
        level    = params['level']
        menu     = params['menu']
        settings = params['settings']
        template = params['template']
        to_base  = self.level2dir(level)
        urldir   = params['urldir']
        parser   = settings['parser']

        soup = BeautifulSoup(template, parser)

        # generate main menu, using info from params['menu']

        main_menu = soup.find("div", id="main-menu")
        ul = soup.new_tag("ul")
        ul['class'] = "clearfix menu sf-menu"
        ul['id'] = "sf-main-menu"
        ul.append('\n')
        for item in menu:
            li = soup.new_tag("li")
            li['class'] = "menu-item"
            ul.append(li)
            ul.append('\n')
            a = soup.new_tag("a", href=item[0]+'/index.html', title=item[1])
            a['wdcms_name'] = item[0]
            if item[0] == "/":
                name = "home"
                a['href'] = "/index.html"
                a['wdcms_name'] = "/."
            else:
                name = os.path.basename(item[0])
            a.append(self.esclinkname(name))
            li.append(a)
        main_menu.append(ul)
        main_menu.append('\n')

        # handle superfish if so desired:
        if (settings['superfish'] == "yes"):
            self.handle_superfish(soup, params)

        # add specific class to main menu anchors
        def has_wdcms_name(tag): return tag.has_attr("wdcms_name")
        x = urldir
        if len(x) == 0:
            x = '/.'
        else:
            x = '/'+urldir[1:].split('/', 1)[0]

        for item in soup.find_all(has_wdcms_name):
            if x == item['wdcms_name']:
                item['class'] = item.get('class', [])+['main-menu-active']
                break  # there should be only one

        # handle head
        for f in head:
            soup.head.append(BeautifulSoup(f, parser))
            soup.head.append("\n")

        # handle css
        for f in css:
            r = self.level2dir(level-f[0])
            l = f[1]
            p = soup.new_tag("link", href=r+"/"+l)
            p['rel'] = "stylesheet"
            soup.head.append(p)
            soup.head.append("\n")

        # handle js
        for f in js:
            r = self.level2dir(level-f[0])
            l = f[1]
            p = soup.new_tag("script", src=r+"/"+l)
            soup.head.append(p)
            soup.head.append("\n")

        # left navigation
        tag = soup.find('div', id='leftlinkstop')

        # find all directories on the same level
        ls = os.listdir('.')
        ls.sort()
        found = False
        for dir in ls:
            if dir != "wdcms.copy" and os.path.isdir(dir):
                found = True
                a = soup.new_tag("a", href=dir+"/index.html")
                a.string = self.esclinkname(dir)
                tag.append(a)
                tag.append(soup.new_tag("br"))
                tag.append('\n')
        if found:
            tag = soup.find('div', id='leftlinkstop_header')
            tag.string = self.esclinkname(os.path.basename(os.getcwd()))

        # create links to children of parent
        # but only if level >=1
        # and Carla allows it.
        if False:  # she doesn't
            if level >= 1:
                tag = soup.find('div', id='leftlinksmiddle')
                ls = os.listdir('./..')
                ls.sort()
                found = False
                for dir in ls:
                    if dir != "wdcms.copy" and os.path.isdir('./../'+dir):
                        found = True
                        a = soup.new_tag("a", href="./../"+dir)
                        name = os.path.basename(os.path.abspath('./../'+dir))
                        a.string = self.esclinkname(name)
                        tag.append(a)
                        tag.append(soup.new_tag("br"))
                        tag.append('\n')
                if found:
                    tag = soup.find('div', id='leftlinksmiddle_header')
                    tag.string = self.esclinkname(
                        os.path.basename(os.path.abspath('./..')))

        # create links to parent's parent directories
        tag = soup.find('div', id='leftlinksbot')

        # find all directories leading to this
        dir = './..'
        found = False
        for i in range(level-1):
            found = True
            a = soup.new_tag("a", href=dir+"/index.html")
            a.string = self.esclinkname(os.path.basename(os.path.abspath(dir)))
            tag.append(a)
            tag.append(soup.new_tag("br"))
            tag.append('\n')
            dir = dir + '/..'
        if found:
            tag = soup.find('div', id='leftlinksbot_header')
            tag.string = "_____"

        # content
        htmlcontent = BeautifulSoup(
            '<div class="markdown">\n'+self.do_format(content, params)+'\n</div>\n' +
            self.copyright, parser)
        # put htmlcontent in <div> with id 'content'
        soup.find('div', id='content').append(htmlcontent)

        page = str(soup)

        # perform translations
        trans = params['trans']
        for key in trans:
            page = page.replace(key, trans[key])
        page = re.sub(self.leftq+r'(.*?)'+self.rightq, r'\1', page)
        page = re.sub(self.leftqq+r'(.*?)'+self.rightqq,
                      self.leftq+r'\1'+self.rightq, page)

        soup = BeautifulSoup(page, parser)

        # adjust hrefs and srcs starting with '/':
        rep = self.repattr()
        for attr in ['href', 'src']:
            rep.doit(to_base, soup, attr)

        page = str(soup)
        open(self.outputdir+urldir+"/index.html",
             "w", encoding=self.encoding).write(page)

    def handle_superfish(self, soup, params):
        def has_wdcms_name(tag): return tag.has_attr("wdcms_name")
        for item in soup.find_all(has_wdcms_name):
            ref    = item['wdcms_name']
            level  = params['level']
            toroot = self.level2dir(level)
            curdir = os.getcwd()
            os.chdir(toroot+'/'+ref)

            ls = os.listdir('.')
            ls.sort()
            ul = soup.new_tag("ul")
            ul.append('\n')
            for f in ls:
                if os.path.isdir(f) and not fnmatch.fnmatch(f, 'wdcms.copy'):
                    li          = soup.new_tag("li")
                    li['class'] = "menu-item"
                    a = soup.new_tag("a", href=ref+'/'+f+"/index.html")
                    a['class'] = "sf-dropdown"
                    a.string   = self.esclinkname(f)
                    li.append(a)
                    ul.append(li)
                    li.insert_after('\n')
            item.insert_after(ul)
            ul.insert_before('\n')
            os.chdir(curdir)

    def recur(self, params):
        params = copy.deepcopy(params)
        self.doitdir(params)
        urldir = params['urldir']
        level = params['level']
        os.makedirs(self.outputdir + "/" + urldir, exist_ok=True)
        ls = os.listdir('.')
        ls.sort()
        print("Visiting:", self.sourceroot+urldir)
        for f in ls:

            if os.path.isdir(f):
                if fnmatch.fnmatch(f, 'wdcms.copy'):
                    lsc = os.listdir(f)
                    for fc in lsc:
                        p = "wdcms.copy/"+fc
                        if os.path.isfile(p):
                            self.numcopy += 1
                            shutil.copy(p, self.outputdir+"/"+urldir+"/"+fc)
                        if os.path.isdir(p):
                            self.numcopy += 1
                            shutil.copytree(p, self.outputdir+"/"+urldir+"/"+fc, 
                                            dirs_exist_ok=True)

            if os.path.isfile(f):
                if (fnmatch.fnmatch(f, 'template.html')):
                    params['template'] = open(
                        f, 'r', encoding=self.encoding).read()
                if fnmatch.fnmatch(f, 'head.html'):
                    self.numhead += 1
                    params['head'].append(
                        open(f, "r", encoding=self.encoding).read())
                for x in "css", "js":
                    if fnmatch.fnmatch(f, '*.'+x):
                        shutil.copy(self.startdir+urldir+"/"+f,
                                    self.outputdir+"/"+urldir)
                        if x == "css":
                            self.numcss += 1
                        if x == "js":
                            self.numjs += 1
                        params[x] += [[level, f]]
                if fnmatch.fnmatch(f, 'content.txt'):
                    params['content'] = open(
                        f, 'r', encoding=self.encoding).read()
                if fnmatch.fnmatch(f, 'trans.txt'):
                    self.handle_translations(params, open(
                        f, 'r', encoding=self.encoding).readlines())
                    self.numtrans += 1
                if fnmatch.fnmatch(f, 'settings.txt'):
                    self.handle_settings(params, open(
                        f, 'r', encoding=self.encoding).readlines())
                    self.numsettings += 1
                if fnmatch.fnmatch(f, 'menu.txt'):
                    self.handle_menu(params, open(
                        f, 'r', encoding=self.encoding).readlines())
                    self.nummenu += 1

        if self.maxfork < 2:                      # no multiprocessing
            self.multi = False
            self.handle_content(params)
            self.nummd += 1
        else:
            self.multi = True
            while True:                           # multiprocessing
                if (self.forked < self.maxfork):  # number of forks < max
                    pid = os.fork()
                    if pid == 0:                  # child code
                        self.handle_content(params)
                        sys.exit(0)
                    else:                         # parent code
                        self.forked += 1
                        self.nummd  += 1
                        break
                else:                             # too many forks, wait until a process is ready
                    pid, rc = os.wait()
                    if (rc != 0):
                        print("Halting because of error",rc>>8,"in forked process");
                        sys.exit(1);
                    self.forked -= 1

        ls1 = ls
        if self.mode:
            if self.mode == '.' and level >= len(self.dirs)-1:
                return
            if level < len(self.dirs)-1:
                ls1 = [self.dirs[level+1]]

        for f in ls1:
            if not os.path.islink(f) and os.path.isdir(f) and f != "wdcms.copy":
                curdir = os.getcwd()
                os.chdir(f)
                params['level'] += 1
                params['urldir'] = urldir + "/"+f
                self.recur(params)
                os.chdir(curdir)
                params['level'] -= 1

    # find start of the website, marked with the file wdcms.root and template.html

    def findstart(self):
        found = False
        toroot = "."
        self.dirs = []
        text = "Searching for root of website (containing 'wdcms.root' and 'template.html)':\n"
        prevdir = ''
        depth = 0
        while True:
            depth += 1
            thisdir = os.getcwd()
            self.dirs = [os.path.basename(thisdir)] + self.dirs
            ls = os.listdir('.')
            text += "examining: "+thisdir+'\n'
            hits = 0
            for f in ls:
                if os.path.isfile(f):
                    if fnmatch.fnmatch(f, 'wdcms.root'):
                        hits += 1
                    if fnmatch.fnmatch(f, 'template.html'):
                        hits += 1
                    if hits == 2:
                        found = True
                        break
            if found or thisdir == prevdir or depth > 1000:
                break
            prevdir = thisdir
            os.chdir('../')
            toroot += "/.."
        if not found:
            print(text)
        self.sourceroot = os.path.basename(os.getcwd())
        return found, toroot

    def start(self):
        self.startdir = os.getcwd()
        print("Creating website")
        print("input starts in '"+os.path.abspath(self.startdir)+"'")
        x = open("wdcms.root", "r", encoding=self.encoding).readlines()
        self.outputdir = "/tmp/website"
        par = -1
        for line in x:
            k, v = self.splitline(line, ':')
            if k == "outputdir":
                self.outputdir = v
            elif k == "parallel":
                try:
                    par = int(v)
                except:
                    pass

        if hasattr(os, "fork") and callable(getattr(os, "fork")) :
            cpus = multiprocessing.cpu_count()
            if par < 0:
                self.maxfork = multiprocessing.cpu_count()
            else:
                self.maxfork = par
        else:
            self.maxfork = 0

        print("Global settings from 'wdcms.root':");
        print("output to:", "'"+self.outputdir+"'")
        print("parallel :", self.maxfork);

        if os.path.isfile("settings.txt"):
            self.handle_settings(self.startparams, open(
                "settings.txt", 'r', encoding=self.encoding).readlines())

        print("Start settings from 'settings.txt':")
        print("parser   :", self.startparams['settings']['parser'])
        print("format   :", self.startparams['settings']['format'])
        print("superfish:", self.startparams['settings']['superfish'])

        if False:  # something like this is needed if absolute urls are demanded
            # localhost:
            # outputbase = "/~willem/websites/easyhosting/httpdocs/ratrabbit"
            # jukebox:
            outputbase = "/ratrabbit"
            # easyhosting:
            # outputbase = "/ratrabbit"

        self.recur(self.startparams)

        if self.multi:  # wait for forked processes to end
            while True:
                try:
                    os.wait()
                except:
                    break

def main():

    arg = None

    if len(sys.argv) >= 2:
        arg = sys.argv[1]
        if arg not in [".", "/"]:
            print("usage: "+sys.argv[0]+" [arg]")
            print("where arg = . or arg = /")
            print("Examples:")
            print("  "+sys.argv[0])
            print("  "+sys.argv[0]+" /")
            sys.exit(1)

    www = wdcms(arg=arg)

    print("wdcms.py version", VERSION)

    if not www.findstart()[0]:
        print("No directory containing 'wdcms.root' and 'template.html' found, quitting.")
        sys.exit(1)

    www.start()

    print("Generated:")
    print("   index.html   :", www.nummd)
    print("Copied:")
    print("   folders      :", www.numdirs)
    print("   *.css        :", www.numcss)
    print("   *.js         :", www.numjs)
    print("   copy         :", www.numcopy)
    print("Used:")
    print("   content.txt  :", www.nummd)
    print("   trans.txt    :", www.numtrans)
    print("   settings.txt :", www.numsettings)
    print("   menu.txt     :", www.nummenu)
    print("   head.html    :", www.numhead)

if __name__ == "__main__":
    main()
