#!/usr/bin/env python3
# LICENSE: MIT
# do whatever you like with this code, all at your own risk
# docs: https://www.ratrabbit.nl/ratrabbit/wdcms
"""
   wdcmsp: a generator for a static web
"""
import copy
import datetime
import fnmatch
import multiprocessing
import os
import re
import shutil
import sys
from subprocess import Popen, PIPE
import platform
import markdown
import markdown2
from bs4 import BeautifulSoup


# ChangeLog
# 0.9.8   initial release
# 0.9.9   add encoding=self.encoding at every (F)open statement
#         define self.encoding as "utf-8"
#         add ChangeLog
#         add self.copyright
#         use shutil.copytree in stead of distutils.dir_util.copy_tree
#         add multiprocessing using os.fork() if available
# 0.9.9.1 add code to wait for end of all forked processes
# 0.9.9.2 adapt to be used in pip or setup.py
# 0.9.9.3 add cmark as formatter
# 0.9.9.4 rearrange code
#         remove "parser" from settings
#         fix bug in handle_superfish
#         add "tidy" option in settings
#         use multiprocessing Pool in stead of fork
# 0.9.9.5 put <br> inside <a> in left pane

# global constants
LEFTQ: str = "YYY"
LEFTQQ: str = "yyy"
RIGHTQ: str = "ZZZ"
RIGHTQQ: str = "zzz"
VERSION: str = "0.9.9.6"
encoding: str = "utf-8"
havepool: bool = hasattr(multiprocessing, "Pool") and callable(getattr(multiprocessing, "Pool"))
parser: str = "html.parser"

class Constants():
    """
    things not changed by handle_dir()
    """
    def __init__(self):
        self.dirs: list[str] = [] # directory names, from the directory
        #                           where wdcms is started to startdir
        self.maxfork: int = 0     # max number of parallel processes
        self.mode: str = ""       # command line parameter
        self.outputdir: str = ""  # output to this directory: root of website
        self.req    = None        # regex for hendling YYY...ZZZ
        self.reqq  = None         # regex for handling yyy...zzz
        self.sourceroot: str = "" # name of the direcory that contains wdcms.root and template.html
        self.startdir: str = ""   # name of directory where wdcms is called
        self.copyright: str = "<!-- Generated by wdcms/python version "+\
                               version()+" Copyright 2021, 2022, 2023 W. Vermin -->"

class Counters():
    """
    counters
    """
    def __init__(self):
        self.forked: int = 0
        self.nummd: int = 0
        self.numdirs: int = 0
        self.numcss: int = 0
        self.numjs: int = 0
        self.numcopy: int = 0
        self.numtrans: int = 0
        self.numsettings: int = 0
        self.nummenu: int = 0
        self.numhead: int = 0

class Page():
    """
    for a page, things that change during traversing by handle_dir()
    """
    def __init__(self):
        self.content: str = ""           # content, in general a markdown page
        self.css: list[(int, str)] = []  # list of pevious .css files
        self.cwd: str = ""               # current directory in input
        self.head: list[str] = []        # list of addendums for <head>
        self.js: list[(int,str)] = []    # list of previous .js files
        self.level: int = 0              # recursion level
        self.menu: list[(str,str)] = []  # menu
        self.multi: bool = False         # multiprocessing
        self.settings: dict[str] = {}    # settings
        self.template: str = ""          # template
        self.trans: dict[str] = {}       # translations
        self.urldir: str = ""            # current directory in web

    def clone(self):
        """
        clone this class
        """
        cloned = Page()

        cloned.content = self.content
        cloned.css = self.css.copy()
        cloned.cwd = self.cwd
        cloned.head = self.head.copy()
        cloned.js = self.js.copy()
        cloned.level = self.level
        cloned.menu = self.menu.copy()
        cloned.multi = self.multi
        cloned.settings = self.settings.copy()
        cloned.template = self.template
        cloned.trans = self.trans.copy()
        cloned.urldir = self.urldir
        return cloned

def do_format(soup: BeautifulSoup, page: Page, constants: Constants) -> BeautifulSoup:
    """
    run the formatter (cmark, pandoc, ...)
    """
    formatter = page.settings['format']
    data = copy.deepcopy(page.content)
    a = data.split('\n', 1)
    if len(a) == 2:
        k, v = splitline(a[0], ':')
        if k == "format":
            formatter = v
            data = a[1]

    if formatter == "markdown":
        html = markdown.markdown(data)
    elif formatter == "markdown2":
        html = markdown2.markdown(data)
    elif formatter == "pandoc":
        p = Popen(["pandoc", "-f", "markdown-smart-auto_identifiers", "-t", "html"],
                  text=True, stdin=PIPE, stdout=PIPE, encoding=encoding)
        # -smart: disable smart quotes
        # -auto_identifiers: disable generation of id in <h> <h1> ...
        out = p.communicate(input=data)
        html = out[0]
    elif formatter == "cmark":
        p = Popen(["cmark", "--unsafe", "-t", "html"],
                  text=True, stdin=PIPE, stdout=PIPE, encoding=encoding)
        out = p.communicate(input=data)
        html = out[0]
    elif formatter == "mediawiki":
        p = Popen(["pandoc", "-f", "mediawiki-smart-auto_identifiers", "-t", "html"],
                  text=True, stdin=PIPE, stdout=PIPE, encoding=encoding)
        # -smart: disable smart quotes
        # -auto_identifiers: disable generation of id in <h> <h1> ...
        out = p.communicate(input=data)
        html = out[0]
    elif formatter == "asis":
        html = data
    else:
        print("Unknown formatter defined in", page['urldir']+': ', formatter +
              ". Valid formatters are 'pandoc' 'cmark' 'markdown' 'markdown2' 'asis' 'mediawiki'")
        sys.exit(1)

    s = "<div class=\"markdown\">\n" + html + "\n</div>\n" + constants.copyright
    htmlcontent = BeautifulSoup(s,parser)
    soup.find('div', id='content').append(htmlcontent)

    # perform translations
    txt = str(soup)

    trans = page.trans
    for key in trans:
        txt = txt.replace(key, trans[key])
    txt = re.sub(LEFTQ+r'(.*?)'+RIGHTQ, r'\1', txt)
    txt = re.sub(LEFTQQ+r'(.*?)'+RIGHTQQ,
                 LEFTQ+r'\1'+RIGHTQ, txt)

    soup = BeautifulSoup(txt, parser)

    # adjust hrefs and srcs starting with '/':
    to_base = level2dir(page.level)
    class Repattr:
        """
        ad-hoc thingy
        """
        def __init__(self):
            self.attr = None
        def has_attr(self, tag):
            """
            too little to say anything abot it
            """
            return tag.has_attr(self.attr)

        def doit(self, to_base, soup, attr):
            """
            too little to say anything abot it
            """
            self.attr = attr
            for child in soup.find_all(self.has_attr):
                x = child[attr]
                if len(x) >= 1 and x[0] == '/':
                    if len(x) >= 2 and x[1] == '/':
                        child[attr] = x[1:]
                    else:
                        child[attr] = to_base + x
    # end of Repattr

    rep = Repattr()
    for attr in ['href', 'src']:
        rep.doit(to_base, soup, attr)

    return soup
# end do_format

def esclinkname(l: str)-> str:
    """
    same as linkname, with _ replaced by space
    """
    return rep_(LEFTQ+l+RIGHTQ)

def findstart(page: Page, constants: Constants)->(bool, Page, list[str]):
    """
    find start of the website, marked with the file wdcms.root and template.html
    """
    found = False
    dirs: list[str] = []
    texts: list[str] = []
    texts += ["Searching for root of website (containing 'wdcms.root' and 'template.html)':"]
    prevdir:str  = ''
    depth:int  = 0
    while True:
        depth += 1
        thisdir = os.getcwd()
        texts += [" examining: " + thisdir]
        dirs = [os.path.basename(thisdir)] + dirs
        found = is_readable_file("wdcms.root") and is_readable_file("template.html")
        if found or thisdir == prevdir or depth > 1000:
            break
        prevdir = thisdir
        os.chdir("../")
    if not found:
        return False, page, texts
    constants.sourceroot = os.path.basename(os.getcwd())
    constants.dirs = dirs
    return True, page, texts
# end of findstart

def getdate() -> str:
    """
    get date
    """
    date = datetime.datetime.now()
    date = date.strftime("%Y/%m/%d %H:%M")  # alas, %Z shows nothing ...
    return date

def handle_dir(pool, presults, page: Page, constants: Constants, counters: Counters):
    """
    recursive function, visiting the input directories
    """
    counters.numdirs += 1
    urldir = page.urldir
    level = page.level
    page.cwd = os.getcwd()
    d = constants.outputdir
    d += "/"
    d += urldir
    try:
        os.makedirs(d, exist_ok=True)
    except:
        print("Cannot create dir: '"+d+"'. Quitting")
        sys.exit(1)

    ls = os.listdir('.')
    ls.sort()

    visit = constants.sourceroot
    visit += urldir
    print("Visiting:", visit)

    f = 'wdcms.copy'
    if is_readable_dir(f):
        lsc = os.listdir(f)
        for fc in lsc:
            p = "./wdcms.copy/"+fc
            if is_readable_file(p):
                counters.numcopy += 1
                shutil.copy(p, constants.outputdir+"/"+urldir+"/"+fc)
            if is_readable_dir(p):
                counters.numcopy += 1
                shutil.copytree(p, constants.outputdir+"/"+urldir+"/"+fc,
                                dirs_exist_ok=True)

    f = 'template.html'
    if is_readable_file(f):
        page.template = open(
            f, 'r', encoding=encoding).read()

    f = 'head.html'
    if is_readable_file(f):
        counters.numhead += 1
        page.head.append(
            open(f, "r", encoding=encoding).read())

    f = 'content.txt'
    if is_readable_file(f):
        page.content = open(
            f, 'r', encoding=encoding).read()

    f = 'trans.txt'
    if is_readable_file(f):
        handle_translations(page, open(
            f, 'r', encoding=encoding).readlines())
        counters.numtrans += 1

    f = 'settings.txt'
    if is_readable_file(f):
        handle_settings(page, open(
            f, 'r', encoding=encoding).readlines())
        counters.numsettings += 1

    f = 'menu.txt'
    if is_readable_file(f):
        page = handle_menu(page, open(
            f, 'r', encoding=encoding).readlines())
        counters.nummenu += 1

    ls1 = ls

    for f in ls:
        for x in "css", "js":
            if fnmatch.fnmatch(f, '*.'+x):
                shutil.copy(constants.startdir+urldir+"/"+f,
                            constants.outputdir+"/"+urldir)
                if x == "css":
                    counters.numcss += 1
                    page.css += [(level, f)]
                if x == "js":
                    counters.numjs += 1
                    page.js += [(level, f)]

    counters.nummd += 1

    results = presults + []

    if (not havepool) or (constants.maxfork < 2):                      # no multiprocessing
        page.multi = False
        handle_template(page,constants)

    else:
        if True:
            page.multi = True
            results.append(pool.apply_async(handle_template,[page.clone(),constants]))

    if constants.mode:
        if constants.mode == '.' and page.level >= len(constants.dirs)-1:
            return results
        if page.level < len(constants.dirs)-1:
            ls1 = [constants.dirs[level+1]]

    for f in ls1:
        if not os.path.islink(f) and is_readable_dir(f) and f != "wdcms.copy":
            curdir = os.getcwd()
            os.chdir(f)
            page.level += 1
            page.urldir = urldir + "/"+f
            page.cwd = os.getcwd()
            results = handle_dir(pool, results + [], page.clone(), constants, counters)
            os.chdir(curdir)
            page.cwd = curdir
            page.level -= 1
    return results
# end handle_dir

def handle_menu(page: Page, lines: list[str]) -> Page:
    """
    handles lines from menu.txt: defining to row menu items
    """
    firstmenu = True
    for line in lines:
        if len(line) > 0 and line[0] != "#":
            k, v = splitline(line, ':')
            if len(v) > 0:
                if k == 'menu':
                    if firstmenu:
                        firstmenu = False
                        page.menu = []
                    l, w = splitline(v, ':')
                    if len(l) > 0:
                        page.menu += [(l, w)]
    return page

def handle_settings(page: Page, lines):
    """
    handle lines from settings.txt
    """
    for line in lines:
        if len(line) > 0 and line[0] != "#":
            k, v = splitline(line, ':')
            if len(v) > 0:
                if k in ['format', 'superfish', 'tidy']:
                    page.settings[k] = v
                else:
                    print("Invalid settings in",
                          './'+page['urldir'], ": '"+k+"', quitting")
                    sys.exit(1)
    return page

def handle_superfish(level: int, name: str)->BeautifulSoup:
    """
    add superfisch things
    """
    curdir = os.getcwd()
    d = level2dir(level)
    d += "/" + name
    os.chdir(d)
    ls = os.listdir('.')
    ls.sort()
    lis = []
    s = BeautifulSoup("<div></div",parser)
    for f in ls:
        if is_readable_dir(f) and f != "wdcms.copy":
            href = name + "/" + f + "/index.html"
            a = s.new_tag("a",href=href)
            a.append(esclinkname(f))
            a["class"] = "sf-dropdown"
            li = s.new_tag("li")
            li["class"] = "menu-item"
            li.append(a)
            lis += ["\n"]
            lis += [li]
    ul = s.new_tag("ul")
    for item in lis:
        ul.append(item)

    os.chdir(curdir)
    return ul
# end of handle_superfish

def handle_template(page: Page, constants: Constants) -> BeautifulSoup:
    """
    This function creates a dom (soup) with the menus in place
    starting from template.html
    and writes this to the desired location
    """
    css      = page.css
    head     = page.head
    js       = page.js
    level    = page.level
    menu     = page.menu
    settings = page.settings
    template = page.template
    urldir   = page.urldir

    os.chdir(page.cwd)

    soup = BeautifulSoup(template, parser)

    # generate main menu, using info from params['menu']

    main_menu = soup.find("div", id="main-menu")
    lis = []
    first = True
    for item in menu:
        href = item[0]
        href += "/index.html"
        name = ""
        wdcms_name = item[0]
        if item[0] == "/":
            name = "home"
            href = "/index.html"
            wdcms_name = "/."
        else:
            name = os.path.basename(item[0])

        name = esclinkname(name)

        # add specific class to main menu anchors
        class_ = "link"
        if first:
            x = urldir
            if len(x) == 0:
                x = "/."
            else:
                x = '/'+urldir[1:].split('/', 1)[0]
            if x == wdcms_name:
                class_ += " main-menu-active"
                first = False
        #
        a = soup.new_tag("a", href=href, title=item[1])
        a["class"] = class_
        a["wdcms_name"] = wdcms_name
        a.append(name)

        lis += ["\n"]
        a = [a]

        # handle superfish if so desired:
        if settings['superfish'] == "yes":
            s = handle_superfish(level, wdcms_name)
            a += [s]
        li = soup.new_tag("li")
        li["class"] = "menu-item"
        for jtem in a:
            li.append(jtem)
        lis += [li]
        lis += ["\n"]
    #
    lis += ["\n"]
    ul = soup.new_tag("ul",id="sf-main-menu")
    ul["class"] = "clearfix menu sf-menu"
    for item in lis:
        ul.append(item)
    main_menu = soup.find("div", id="main-menu")
    main_menu.append(ul)
    main_menu.append('\n')

    # handle head
    for f in head:
        soup.head.append(BeautifulSoup(f, parser))
        soup.head.append("\n")

    # handle css
    for f in css:
        r = level2dir(level-f[0])
        l = f[1]
        p = soup.new_tag("link", href=r+"/"+l)
        p['rel'] = "stylesheet"
        soup.head.append(p)
        soup.head.append("\n")

    # handle js
    for f in js:
        r = level2dir(level-f[0])
        l = f[1]
        p = soup.new_tag("script", src=r+"/"+l)
        soup.head.append(p)
        soup.head.append("\n")

    # left navigation
    tag = soup.find('div', id='leftlinkstop')

    # find all directories on the same level
    ls = os.listdir('.')
    ls.sort()
    found = False
    for item in ls:
        if item != "wdcms.copy" and os.path.isdir(item):
            found = True
            a = soup.new_tag("a", href=item+"/index.html")
            a.string = esclinkname(item)
            a.append(soup.new_tag("br"))
            tag.append(a)
            #tag.append(soup.new_tag("br"))
            tag.append('\n')
    if found:
        tag = soup.find('div', id='leftlinkstop_header')
        tag.string = esclinkname(os.path.basename(os.getcwd()))

    # create links to children of parent
    # but only if level >=1
    # and Carla allows it.
    if True is False:  # she doesn't
        if level >= 1:
            tag = soup.find('div', id='leftlinksmiddle')
            ls = os.listdir('./..')
            ls.sort()
            found = False
            for item in ls:
                if item != "wdcms.copy" and os.path.isdir('./../'+item):
                    found = True
                    a = soup.new_tag("a", href="./../"+item)
                    name = os.path.basename(os.path.abspath('./../'+item))
                    a.string = esclinkname(name)
                    tag.append(a)
                    tag.append(soup.new_tag("br"))
                    tag.append('\n')
            if found:
                tag = soup.find('div', id='leftlinksmiddle_header')
                tag.string = esclinkname(
                    os.path.basename(os.path.abspath('./..')))

    # create links to parent's parent directories
    tag = soup.find('div', id='leftlinksbot')

    # find all directories leading to this
    mydir = './..'
    found = False
    for _i in range(level-1):
        found = True
        a = soup.new_tag("a", href=mydir+"/index.html")
        a.string = esclinkname(os.path.basename(os.path.abspath(mydir)))
        tag.append(a)
        tag.append(soup.new_tag("br"))
        tag.append('\n')
        mydir = mydir + '/..'
    if found:
        tag = soup.find('div', id='leftlinksbot_header')
        tag.string = "_____"
    #
    soup = do_format(soup,page,constants)
    write_page(soup, page, constants)
# end of handle_template

def handle_translations(page: Page, lines: []) -> Page:
    """
    handles lines from trans.txt
    """
    for line in lines:
        if len(line) > 0 and line[0] != "#":
            k, v = splitline(line, ':')
            if len(v) > 0:
                page.trans[linkname(k)] = copy.copy(v)
                q = k.find("_")
                if q >= 0:
                    k = esclinkname(k)
                    page.trans[linkname(k)] = copy.copy(v)
    return page

def init(page: Page, constants: Constants, i: str) -> Page:
    """
    initialize
    """
    constants.mode = i
    lines = [
        "wdcms_date: "+ getdate(),
        "wdcms_version: " + version(),
        "leftq: " + LEFTQQ,
        "rightq: " + RIGHTQQ
    ]
    page = handle_translations(page, lines)

    lines = [
        "superfish: yes",
        "format: pandoc",
        "tidy:   yes",
    ]
    page = handle_settings(page,lines)

    rc, page, texts = findstart(page, constants)
    if not rc:
        for t in texts:
            print(t)
        print("No directory containing 'wdcms.root' and 'template.html' found, quitting.")
        sys.exit(1)
    return page
# end of init

def is_readable_dir(dirname: str) -> bool:
    """
    determine if dirname is a readable directory (sloppy implementation)
    """
    if os.path.isdir(dirname):
        oldwd = os.getcwd()
        try:
            os.chdir(dirname)
            os.chdir(oldwd)
            return True
        except:
            os.chdir(oldwd)
    return False

def is_readable_file(filename: str) -> bool:
    """
    determine if filename is a readable file (sloppy implementation)
    """
    if os.path.isfile(filename):
        try:
            f = open(filename,"r")
            f.close()
            return True
        except:
            pass
    return False

def level2dir(level):
    """
    translate level to path to root of web
    """
    d = '.'
    for _i in range(level):
        d += '/..'
    return d

def linkname(l: str)->str:
    """
    returns l, with LEFTQ and RIGHTQ
    """
    return LEFTQ+l+RIGHTQ

def main():
    """
    the main progam
    """
    arg = None
    print("wdcmsp.py version", version())
    print("Detected system type:      "+platform.system())
    if havepool:
        print("Number of (virtual) cpu's: ", multiprocessing.cpu_count())

    page = Page()
    counters = Counters()
    constants = Constants()

    if len(sys.argv) >= 2:
        arg = sys.argv[1]
        if arg not in [".", "/"]:
            print("usage: "+sys.argv[0]+" [arg]")
            print("where arg = . or arg = /")
            print("Examples:")
            print("  "+sys.argv[0])
            print("  "+sys.argv[0]+" /")
            sys.exit(1)

    page = init(page, constants, arg)
    start(page, constants, counters)

    print("Generated:")
    print("   index.html   :", counters.nummd)
    print("Copied:")
    print("   folders      :", counters.numdirs)
    print("   *.css        :", counters.numcss)
    print("   *.js         :", counters.numjs)
    print("   copy         :", counters.numcopy)
    print("Used:")
    print("   content.txt  :", counters.nummd)
    print("   trans.txt    :", counters.numtrans)
    print("   settings.txt :", counters.numsettings)
    print("   menu.txt     :", counters.nummenu)
    print("   head.html    :", counters.numhead)
# end of main

def rep_(a: str) -> str:
    """
    see below
    """
    return a.replace("_", " ").strip()

def splitline(line, c):
    """
    splits line
    """
    l = re.sub(r'#.*$', r'', line)
    l = l.split(c, 1)
    if len(l) == 2:
        return l[0].strip(), l[1].strip()
    return "", ""

def start(page: Page, constants: Constants, counters: Counters):
    """
    this runs the program
    """
    constants.startdir = os.getcwd()
    print("Creating website")
    print("input starts in '"+os.path.abspath(constants.startdir)+"'")
    constants.outputdir = "/tmp/website"
    lines = open("wdcms.root", "r", encoding=encoding).readlines()
    for line in lines:
        k, v = splitline(line, ':')
        if len(v) > 0:
            if k == "outputdir":
                constants.outputdir = v
            elif k == "parallel":
                try:
                    par = int(v)
                except:
                    pass

    if havepool:
        if par < 0:
            constants.maxfork = multiprocessing.cpu_count()
        else:
            constants.maxfork = par
    else:
        constants.maxfork = 0

    print("Global settings from 'wdcms.root':")
    print("   output to:", "'"+constants.outputdir+"'")
    print("   parallel :", constants.maxfork)

    if is_readable_file("settings.txt"):
        handle_settings(page, open(
            "settings.txt", 'r', encoding=encoding).readlines())

    print("Start settings from 'settings.txt':")
    print("   format   :", page.settings['format'])
    print("   superfish:", page.settings['superfish'])
    print("   tidy     :", page.settings['tidy'])

    if havepool:
        pool = multiprocessing.Pool(processes = constants.maxfork)
    else:
        pool = None

    results = []

    results = handle_dir(pool, results, page, constants, counters)

    for r in results:
        x = r.get()
        if x != None:
            print("Unexpected result from handle_template: ",x)

# end of start

def version():
    """
    return version
    """
    return VERSION

def write_page(soup: BeautifulSoup, page: Page, constants: Constants) -> None:
    """
    finally: write the generated page
    """
    out = constants.outputdir
    out += page.urldir
    out += "/index.html"
    html = str(soup)
    if page.settings['tidy'] == "yes":
        p = Popen(["tidy", "-i", "-q", "--show-warnings", "no"],
                  text=True, stdin=PIPE, stdout=PIPE, encoding=encoding)
        x = p.communicate(input=html)
        html = x[0]
    open(out, "w", encoding=encoding).write(html)
#end of write_page

if __name__ == "__main__":
    main()
